import logging
from typing import Optional, TypeVar, Union

# import oic
from oic.extension.client import Client as ClientExtension
from oic.oic.consumer import Consumer
from oic.utils.authn.client import CLIENT_AUTHN_METHOD

from django_pyoidc.exceptions import (
    InvalidOIDCConfigurationException,
    InvalidSIDException,
)
from django_pyoidc.session import OIDCCacheSessionBackendForDjango
from django_pyoidc.settings import OIDCSettings, OIDCSettingsFactory, OidcSettingValue
from django_pyoidc.utils import OIDCCacheBackendForDjango

logger = logging.getLogger(__name__)

T = TypeVar("T")


class OIDCClient:
    def __init__(self, op_name: str, session_id: Optional[str] = None):
        self.opsettings = OIDCSettingsFactory.get(op_name)

        self.session_cache_backend = OIDCCacheSessionBackendForDjango(self.opsettings)
        self.general_cache_backend = OIDCCacheBackendForDjango(self.opsettings)
        client_id = self.opsettings.get("client_id")
        client_secret = self.opsettings.get("client_secret", None)
        consumer_config = self.opsettings.get(
            "client_consumer_config_dict",
            {
                # "debug": True,
                "response_type": "code"
            },
        )

        client_config = {
            "client_id": client_id,
            "client_authn_method": self.opsettings.get(
                "client_authn_method", CLIENT_AUTHN_METHOD
            ),
        }
        self.consumer = Consumer(
            session_db=self.session_cache_backend,
            consumer_config=consumer_config,
            client_config=client_config,
        )  # type: ignore[no-untyped-call] # oic.oic.consumer.Consumer is not typed yet

        # used in token introspection
        self.client_extension = ClientExtension(**client_config)  # type: ignore[no-untyped-call] # oic.extension.client.Client is not typed yet

        provider_discovery_uri: str = self.opsettings.get("provider_discovery_uri", None)  # type: ignore[assignment] # we can assume that the configuration is ok
        self.client_extension.client_secret = client_secret

        if session_id is not None:
            try:
                self.consumer.restore(session_id)  # type: ignore[no-untyped-call] # Consumer.restore is not typed yet
            except KeyError:
                # This is an error as for example during the first communication round trips between
                # the op and the client we'll have to find state elements in the oidc session
                raise InvalidSIDException(
                    f"OIDC consumer failed to restore oidc session {session_id}."
                )
            return

        if provider_discovery_uri is None:
            raise InvalidOIDCConfigurationException(
                "No provider discovery uri provided."
            )
        else:
            if self.opsettings.get("oidc_cache_provider_metadata", False):
                cache_key = self.general_cache_backend.generate_hashed_cache_key(
                    provider_discovery_uri
                )
                try:
                    config = self.general_cache_backend[cache_key]
                    # this will for example register endpoints on the consumer object
                    self.consumer.handle_provider_config(config, provider_discovery_uri)  # type: ignore[arg-type] # provider_discovery_uri is from the cache
                except KeyError:
                    # This make an HTTP call on provider discovery uri
                    config = self.consumer.provider_config(provider_discovery_uri)
                    # shared microcache for provider config
                    ttl: int = self.opsettings.get("oidc_cache_provider_metadata_ttl")  # type: ignore[assignment] # we can assume the configuration is right
                    self.general_cache_backend.set(
                        cache_key,
                        config,
                        ttl,
                    )
            else:
                # This make an HTTP call on provider discovery uri
                self.consumer.provider_config(provider_discovery_uri)
        self.consumer.client_secret = client_secret

    def get_settings(self) -> OIDCSettings:
        return self.opsettings

    def get_setting(
        self, name: str, default: Optional[T] = None
    ) -> Optional[Union[OidcSettingValue, T]]:
        return self.opsettings.get(name, default)



from typing import Any, Dict, Optional, TypedDict

ProviderConfig = TypedDict(
    "ProviderConfig",
    {
        # most important ----
        "client_id": Optional[str],
        "client_secret": Optional[str],
        "oidc_cache_provider_metadata": Optional[str],
        "oidc_paths_prefix": str,
        "oidc_callback_path": str,
        # less important ---
        "provider_discovery_uri": str,
        "oidc_logout_redirect_parameter_name": str,
        # Use introspection in API-Bearer mode?
        "use_introspection_on_access_tokens": bool,
        # Rare usages ---
        "client_authn_method": Optional[bool],
        "oidc_logout_query_string_redirect_parameter": Optional[str],
        "oidc_logout_query_string_extra_parameters_dict": Optional[Dict[str, Any]],
        # "client_consumer_config_dict": None,
        # some providers may return even more stuff (...) ---
    },
)


class Provider:
    """
    This is the base `Provider` class that is used to implement common provider configuration patterns. You should not
    use this class directly. Instead, you should but subclass it to implement the configuration logic.
    """

    def __init__(self, *args: Any, op_name: str, **kwargs: Any):
        """
        Parameters:
            op_name (str): the name of the sso provider that you are using
        """

        self.op_name = op_name

        if "provider_discovery_uri" in kwargs:
            self.provider_discovery_uri = kwargs["provider_discovery_uri"]
        else:
            self.provider_discovery_uri = None

        if "oidc_logout_redirect_parameter_name" in kwargs:
            self.oidc_logout_redirect_parameter_name = kwargs[
                "oidc_logout_redirect_parameter_name"
            ]
        else:
            self.oidc_logout_redirect_parameter_name = "post_logout_redirect"

        if "oidc_paths_prefix" in kwargs:
            self.oidc_paths_prefix = kwargs["oidc_paths_prefix"]
        else:
            self.oidc_paths_prefix = self.op_name
        if "oidc_callback_path" in kwargs:
            self.oidc_callback_path = kwargs["oidc_callback_path"]
        else:
            self.oidc_callback_path = f"{self.oidc_paths_prefix}-callback"

    def get_default_config(self) -> ProviderConfig:
        """Get the default configuration settings for this provider.

        This configuration defaults are used to provide default values for OIDCSettings.
        User can override these defaults by playing with OIDCSettings arguments.
        """
        return ProviderConfig(
            # most important ----
            client_id=None,
            client_secret=None,
            oidc_cache_provider_metadata=None,
            oidc_paths_prefix=self.oidc_paths_prefix,
            oidc_callback_path=self.oidc_callback_path,
            # less important ---
            provider_discovery_uri=self.provider_discovery_uri,
            oidc_logout_redirect_parameter_name=self.oidc_logout_redirect_parameter_name,
            # Use introspection in API-Bearer mode?
            use_introspection_on_access_tokens=self.op_name == "drf",
            # Rare usages ---
            client_authn_method=None,
            oidc_logout_query_string_redirect_parameter=None,
            oidc_logout_query_string_extra_parameters_dict=None,
        )



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
import os
import sys

import django
from django.conf import settings

sys.path.insert(0, os.path.abspath(".."))

# pass settings into configure
settings.configure(
    INSTALLED_APPS=[
        "django.contrib.auth",
        "django.contrib.contenttypes",
        "django.contrib.sessions",
        "django_pyoidc",
    ]
)

# call django.setup to load installed apps and other stuff
django.setup()

copyright = "Free Software"
project = "Makina Django OIDC"
author = "Paul FLORENCE, Régis LEROY (regilero)"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.doctest",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx.ext.autosectionlabel",
    "sphinx_rtd_theme",
]

intersphinx_cache_limit = 14  # cache for 2 weeks
intersphinx_timeout = 3  # 3 seconds timeout
intersphinx_mapping = {
    "python": ("https://docs.python.org/3.10/", None),
    "django": (
        "https://docs.djangoproject.com/en/stable/",
        "https://docs.djangoproject.com/en/stable/_objects/",
    ),
}

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]


def setup(app):
    app.add_crossref_type(
        directivename="setting",
        rolename="setting",
        indextemplate="pair: %s; setting",
    )



import hashlib
import logging
from importlib import import_module
from typing import Any, Dict, Mapping, MutableMapping, Optional, Union

from django.core.cache import BaseCache, caches

from django_pyoidc.exceptions import ClaimNotFoundError
from django_pyoidc.settings import OIDCSettings

logger = logging.getLogger(__name__)


def import_object(
    path: str, def_name: str
) -> Any:  # FIXME : this function is hard to type correctly
    try:
        mod, cls = path.split(":", 1)
    except ValueError:
        mod = path
        cls = def_name

    return getattr(import_module(mod), cls)


def extract_claim_from_tokens(
    claim: str, tokens: Dict[str, Any], raise_exception: bool = True
) -> Any:
    """Given a dictionnary of tokens claims, extract the given claim.

    This function will seek in "info_token_claims", then "id_token_claims"
    and finally "access_token_claims".
    If the claim is not found a ClaimNotFoundError exception is raised.
    """
    if "info_token_claims" in tokens and claim in tokens["info_token_claims"]:
        value = tokens["info_token_claims"][claim]
    elif "id_token_claims" in tokens and claim in tokens["id_token_claims"]:
        value = tokens["id_token_claims"][claim]
    elif "access_token_claims" in tokens and claim in tokens["access_token_claims"]:
        value = tokens["access_token_claims"][claim]
    else:
        if raise_exception:
            raise ClaimNotFoundError(f"{claim} not found in available OIDC tokens.")
        else:
            return None
    return value


def check_audience(client_id: str, access_token_claims: Dict[str, Any]) -> bool:
    """Verify that the current client_id is present in 'aud' claim.

    Audences are stored in 'aud' claim.
    Audiences of an access token is a list of client_id where this token is allowed.
    When receiving an access token in 'API' bearer mode checking that your client_id
    is in the audience is a must.
    Access tokens received in 'full' mode, when this Django is the OIDC client
    managing the redirections to the SSO server may not always contain the client_id
    in 'aud'. This is the case for Keycloak for example, where the 'aud' would only
    contain 'others' client_id where this token can be used, and not the one generating it.
    Audience in userinfo and id tokens are different beasts.
    """
    if "aud" not in access_token_claims:
        return False
    if client_id not in access_token_claims["aud"]:
        # in case we are current requester of the access token the audience may not be presence in
        # 'aud' (that's the case in recent keycloak) but it should then be in azp (requester of the token).
        if "azp" in access_token_claims and client_id == access_token_claims["azp"]:
            return True
        logger.error(
            f"{client_id} not found in access_token_claims['aud']: {access_token_claims['aud']}"
        )
        return False
    return True


class OIDCCacheBackendForDjango:
    """Implement General cache for OIDC using django cache"""

    def __init__(self, opsettings: OIDCSettings):
        self.op_name = opsettings.get("op_name")

        self.enabled = opsettings.get("oidc_cache_provider_metadata", False)
        if self.enabled:
            cache_key: str = opsettings.get("cache_django_backend")  # type: ignore[assignment] # we can assume that the configuration is right
            self.storage: BaseCache = caches[cache_key]

    def generate_hashed_cache_key(self, value: str) -> str:
        h = hashlib.new("sha256")
        h.update(value.encode())
        cache_key = h.hexdigest()
        return cache_key

    def clear(self) -> Optional[int]:
        if self.enabled:
            self.storage.clear()
            return None
        else:
            return 0

    def get_key(self, key: str) -> str:
        return f"oidc-{self.op_name}-{key}"

    def set(self, key: str, value: Mapping[str, Union[str, bool]], expiry: int) -> None:
        if self.enabled:
            self.storage.set(self.get_key(key), value, expiry)

    def __getitem__(self, key: str) -> MutableMapping[str, Union[str, bool]]:
        if self.enabled:
            data = self.storage.get(self.get_key(key))
            if data is None:
                raise KeyError  # Makes __getItem__ handle like Python dict
            return data
        else:
            raise KeyError



import functools
import logging
from typing import Any, Optional, Tuple

from django.core.exceptions import PermissionDenied
from rest_framework import exceptions
from rest_framework.authentication import BaseAuthentication
from rest_framework.request import Request
from typing_extensions import override

from django_pyoidc.client import OIDCClient
from django_pyoidc.engine import OIDCEngine
from django_pyoidc.exceptions import ExpiredToken
from django_pyoidc.settings import OIDCSettings, OIDCSettingsFactory
from django_pyoidc.utils import OIDCCacheBackendForDjango, check_audience

logger = logging.getLogger(__name__)


class OidcAuthException(Exception):
    pass


class OIDCBearerAuthentication(BaseAuthentication):
    @functools.cached_property
    def opsettings(self) -> OIDCSettings:
        return OIDCSettingsFactory.get("drf")

    @functools.cached_property
    def general_cache_backend(self) -> OIDCCacheBackendForDjango:
        return OIDCCacheBackendForDjango(self.opsettings)

    @functools.cached_property
    def engine(self) -> OIDCEngine:
        return OIDCEngine(self.opsettings)

    @functools.cached_property
    def client(self) -> OIDCClient:
        return OIDCClient("drf")

    def extract_access_token(self, request: Request) -> str:
        val = request.headers.get("Authorization")
        if not val:
            msg = "Request missing the authorization header."
            raise OidcAuthException(msg)
        val = val.strip()
        bearer_name, access_token_jwt = val.split(maxsplit=1)
        requested_bearer_name = self.opsettings.get("oidc_api_bearer_name", "Bearer")
        if not bearer_name.lower() == requested_bearer_name.lower():  # type: ignore[union-attr] # we can assume that this setting is a string
            msg = f"Bad authorization header, invalid Keyword for the bearer, expecting {requested_bearer_name} (check setting oidc_api_bearer_name)."
            raise OidcAuthException(msg)
        return access_token_jwt

    @override
    def authenticate(self, request: Request) -> Optional[Tuple[Any, Any]]:
        """
        Returns two-tuple of (user, token) if authentication succeeds,
        or None otherwise.
        """
        try:
            user = None
            access_token_claims = None

            # Extract the access token from an HTTP Authorization Bearer header
            try:
                access_token_jwt = self.extract_access_token(request)
            except OidcAuthException as e:
                logger.debug(e)
                return None

            # This introspection of the token is made by the SSO server
            # so it is quite slow, but there's a cache added based on the token expiration
            # it could also call a user defined validator if 'use_introspection_on_access_tokens'
            # is False. or it could return None if the two previous are not defined.
            try:
                access_token_claims = self.engine.introspect_access_token(
                    access_token_jwt, client=self.client
                )
            except ExpiredToken:
                msg = "Inactive access token."
                raise exceptions.AuthenticationFailed(msg)

            if not access_token_claims:
                exceptions.AuthenticationFailed(
                    "Access token claims failed to be extracted."
                )
            logger.debug(access_token_claims)
            if not access_token_claims.get("active"):
                msg = "Inactive access token."
                raise exceptions.AuthenticationFailed(msg)

            # FIXME: add an option to request userinfo here, but that may be quite slow

            if access_token_claims:
                logger.debug("Request has valid access token.")

                # FIXME: Add a setting to disable
                client_id: str = self.opsettings.get("client_id")  # type: ignore[assignment] # we can assume that client_id is correctly configured
                if not check_audience(client_id, access_token_claims):
                    raise PermissionDenied(
                        f"Invalid result for acces token audiences check for {client_id}."
                    )

                logger.debug("Let application load user via user hook.")
                user = self.engine.call_get_user_function(
                    tokens={
                        "access_token_jwt": access_token_jwt,
                        "access_token_claims": access_token_claims,
                    },
                    client=self.client,
                )

            if not user:
                # OIDC Bearer Authentication process failure : cannot set active authenticated user
                return None

        except PermissionDenied as exp:
            raise exp
        except Exception as exp:
            logger.exception(exp)
            return None

        return user, access_token_claims



import logging

logger = logging.getLogger(__name__)

try:
    from drf_spectacular.extensions import OpenApiAuthenticationExtension

    from django_pyoidc.settings import OIDCSettingsFactory

    class OIDCScheme(OpenApiAuthenticationExtension):  # type: ignore[no-untyped-call] # drf_spectacular.plumbing.OpenApiGeneratorExtension.__init_subclass__ is untyped
        target_class = "django_pyoidc.drf.authentication.OIDCBearerAuthentication"
        name = "openIdConnect"
        match_subclasses = True
        priority = -1

        @classmethod
        def get_security_definition(cls, auto_schema):  # type: ignore[no-untyped-def] # we do not want to type third party libraries
            # from django_pyoidc.drf.authentication import OIDCBearerAuthentication

            opsettings = OIDCSettingsFactory.get("drf")
            well_known_url: str = opsettings.get("provider_discovery_uri")  # type: ignore[assignment] # we can assume that this is an str
            if not well_known_url.endswith(".well-known/openid-configuration"):
                well_known_url = (
                    well_known_url + "/.well-known/openid-configuration"
                )  # well_known_url should not end with a "/" as it is sanitized

            header_name = opsettings.get("oidc_api_bearer_name", "Bearer")
            if header_name != "Bearer":
                logger.warning(
                    "The configuration for 'oidc_api_bearer_name' will cause issue with swagger UI :"
                    "it is not yet possible to change the header name for swagger UI, you should stick to"
                    "'Bearer'."
                )
            return {
                "type": "openIdConnect",
                "openIdConnectUrl": well_known_url,
            }

except ImportError:
    logger.debug(
        "The package 'drf-spectacular' is not installed, skipping schema generation."
    )



from typing_extensions import override

from django_pyoidc.providers.provider import Provider, ProviderConfig


class LemonLDAPng2Provider(Provider):
    """
    Provide Django settings/urlconf based on LemonLDAP-ng behaviour (v2)
    """

    @override
    def get_default_config(self) -> ProviderConfig:
        result = super().get_default_config()
        result["oidc_logout_query_string_extra_parameters_dict"] = {"confirm": 1}
        return result



from unittest import mock

from django.conf import settings
from django.urls import reverse

from tests.utils import OIDCTestCase


class UrlsTestCase(OIDCTestCase):
    def test_reverse_urls_are_defined(self, *args):
        """
        Test that urls defined in our test settings are working.
        """
        self.assertEqual(
            reverse("test_blogout"),
            "/back_channel_logout-xyz/",
        )
        self.assertEqual(
            reverse("test_login"),
            "/login-xyz/",
        )
        self.assertEqual(
            reverse("test_logout"),
            "/logout-xyz/",
        )
        self.assertEqual(
            reverse("test_logout"),
            "/logout-xyz/",
        )

    @mock.patch("django_pyoidc.client.Consumer.provider_config")
    def test_callback_path_with_uri_name(self, *args):
        """
        Test that using callback_uri_name we have the right callback path.

        "sso1" definition is not using oidc_callback_path but a named route
        with callback_uri_name instead.
        """
        host = settings.DJANGO_PYOIDC["sso1"]["login_uris_redirect_allowed_hosts"][0]
        response = self.client.get(
            reverse("test_login"),
            data={"next": f"https://{host}/foo/bar"},
            SERVER_NAME=host,
        )
        self.assertEqual(response.status_code, 302)
        location = response.headers["Location"]
        elements = location.split("?")
        query_string = elements[1]
        arguments = query_string.split("&")
        self.assertIn("client_id=1", arguments)
        self.assertIn(f"redirect_uri=http%3A%2F%2F{host}%2Fcallback-xyz%2F", arguments)

    @mock.patch("django_pyoidc.client.Consumer.provider_config")
    def test_callback_path_with_oidc_callback_path(self, *args):
        """
        Test that using only oidc_callback_path we have the right callback path.

        "sso2" definition does not contain callback_uri_name not oidc_paths_prefix
        but oidc_callback_path instead.
        """
        host = settings.DJANGO_PYOIDC["sso2"]["login_uris_redirect_allowed_hosts"][0]
        response = self.client.get(
            reverse("test_login_2"),
            data={"next": f"https://{host}/foo/bar"},
            SERVER_NAME=host,
        )
        self.assertEqual(response.status_code, 302)
        location = response.headers["Location"]
        elements = location.split("?")
        query_string = elements[1]
        arguments = query_string.split("&")
        self.assertIn("client_id=2", arguments)
        self.assertIn(f"redirect_uri=http%3A%2F%2F{host}%2Fcallback-wtf%2F", arguments)

    @mock.patch("django_pyoidc.client.Consumer.provider_config")
    def test_helper_callback_path_with_callback_uri_name_and_url_prefix(self, *args):
        """ """
        host = settings.DJANGO_PYOIDC["sso10"]["login_uris_redirect_allowed_hosts"][0]
        response = self.client.get(
            reverse("op10_namespace:sso10-login"),
            data={"next": f"https://{host}/foo/bar"},
            SERVER_NAME=host,
        )
        self.assertEqual(response.status_code, 302)
        location = response.headers["Location"]
        elements = location.split("?")
        query_string = elements[1]
        arguments = query_string.split("&")
        self.assertIn("client_id=10", arguments)
        self.assertIn("response_type=code", arguments)
        self.assertIn(f"redirect_uri=http%3A%2F%2F{host}%2Fsso10-callback", arguments)

    @mock.patch("django_pyoidc.client.Consumer.provider_config")
    def test_helper_callback_path_with_oidc_paths_prefix_and_url_prefix(self, *args):
        """ """
        host = settings.DJANGO_PYOIDC["sso11"]["login_uris_redirect_allowed_hosts"][0]
        response = self.client.get(
            reverse("op11_namespace:sso11-login"),
            data={"next": f"https://{host}/foo/bar"},
            SERVER_NAME=host,
        )
        self.assertEqual(response.status_code, 302)
        location = response.headers["Location"]
        elements = location.split("?")
        query_string = elements[1]
        arguments = query_string.split("&")
        self.assertIn("client_id=11", arguments)
        self.assertIn("response_type=code", arguments)
        # No path prefix for sso11, as we do not use a named route or any way to fix the route prefix.
        self.assertIn(f"redirect_uri=http%3A%2F%2F{host}%2Foidc11-callback", arguments)

    @mock.patch("django_pyoidc.client.Consumer.provider_config")
    def test_helper_callback_path_with_oidc_paths_prefix_and_oidc_callback_path_and_url_prefix(
        self, *args
    ):
        """ """
        host = settings.DJANGO_PYOIDC["sso12"]["login_uris_redirect_allowed_hosts"][0]
        response = self.client.get(
            reverse("op12_namespace:sso12-login"),
            data={"next": f"https://{host}/foo/bar"},
            SERVER_NAME=host,
        )
        self.assertEqual(response.status_code, 302)
        location = response.headers["Location"]
        elements = location.split("?")
        query_string = elements[1]
        arguments = query_string.split("&")
        self.assertIn("client_id=12", arguments)
        self.assertIn("response_type=code", arguments)
        self.assertIn(
            f"redirect_uri=http%3A%2F%2F{host}%2Fprefix12%2Foidc12-zz-callback",
            arguments,
        )

    @mock.patch("django_pyoidc.client.Consumer.provider_config")
    def test_helper_allback_path_with_oidc_paths_prefix_and_callback_uri_name_and_route_prefix(
        self, *args
    ):
        """ """
        host = settings.DJANGO_PYOIDC["sso13"]["login_uris_redirect_allowed_hosts"][0]
        response = self.client.get(
            reverse("op13_namespace:sso13-login"),
            data={"next": f"https://{host}/foo/bar"},
            SERVER_NAME=host,
        )
        self.assertEqual(response.status_code, 302)
        location = response.headers["Location"]
        elements = location.split("?")
        query_string = elements[1]
        arguments = query_string.split("&")
        self.assertIn("client_id=13", arguments)
        self.assertIn("response_type=code", arguments)
        self.assertIn(
            f"redirect_uri=http%3A%2F%2F{host}%2Fprefix13%2Foidc13-ww-callback",
            arguments,
        )



from django.apps import AppConfig


class DjangoPyoidcConfig(AppConfig):
    name = "django_pyoidc"

    def ready(self) -> None:
        # Register schema with drf_spectacular
        from .drf import schema  # noqa: F401



import datetime
import logging
from typing import Any, Dict, MutableMapping, Optional, Union

from django_pyoidc import get_user_by_email
from django_pyoidc.client import OIDCClient
from django_pyoidc.exceptions import ExpiredToken, TokenError
from django_pyoidc.settings import OIDCSettings
from django_pyoidc.utils import OIDCCacheBackendForDjango, import_object

logger = logging.getLogger(__name__)


class OIDCEngine:
    def __init__(self, opsettings: OIDCSettings):
        self.opsettings = opsettings
        self.general_cache_backend = OIDCCacheBackendForDjango(opsettings)

    def call_function(self, setting_func_name: str, *args: Any, **kwargs: Any) -> Any:
        function_path: Optional[str] = self.opsettings.get(setting_func_name)  # type: ignore[assignment] # we can assume that the configuration is right
        if function_path is not None:
            func = import_object(function_path, "")
            return func(*args, **kwargs)

    def call_get_user_function(
        self, client: OIDCClient, tokens: Optional[Dict[str, Any]] = None
    ) -> Any:
        if tokens is None:
            tokens = {}
        if self.opsettings.get("hook_get_user") is not None:
            logger.debug("OIDC, Calling user hook on get_user")
            return self.call_function("hook_get_user", client=client, tokens=tokens)
        else:
            logger.debug("OIDC, Calling get_user_by_email")
            return get_user_by_email(tokens)

    def introspect_access_token(
        self, access_token_jwt: Optional[str], client: OIDCClient
    ) -> Any:
        """
        Perform a cached introspection call to extract claims from encoded jwt of the access_token
        """
        # FIXME: allow a non-cached mode by global settings

        if access_token_jwt is None:
            raise TokenError("Nothing in access_token_jwt.")

        if self.opsettings.get("use_introspection_on_access_tokens"):
            return self._call_introspection(access_token_jwt, client)
        else:
            return self.call_validate_tokens_hook(access_token_jwt, client)

    def _call_introspection(
        self, access_token_jwt: str, client: OIDCClient
    ) -> MutableMapping[str, Union[str, bool]]:
        cache_key = self.general_cache_backend.generate_hashed_cache_key(
            access_token_jwt
        )
        try:
            access_token_claims = self.general_cache_backend[cache_key]
        except KeyError:
            # CACHE MISS

            # RFC 7662: token introspection: ask SSO to validate and render the jwt as json
            # this means a slow http call
            request_args = {
                "token": access_token_jwt,
                "token_type_hint": "access_token",
            }
            client_auth_method = client.consumer.registration_response.get(
                "introspection_endpoint_auth_method", "client_secret_basic"
            )  # type: ignore[no-untyped-call] # oic is untyped
            introspection = client.client_extension.do_token_introspection(
                request_args=request_args,
                authn_method=client_auth_method,
                endpoint=client.consumer.introspection_endpoint,  # type: ignore
            )
            access_token_claims = introspection.to_dict()
            if "active" in access_token_claims and not access_token_claims["active"]:
                # there will not be other claims, like expiry, this is simply an expired token
                logger.info("access token introspection failed, expired token.")
                raise ExpiredToken("Inactive access token.")
            # store it in cache
            current = datetime.datetime.now().strftime("%s")
            if "exp" not in access_token_claims:
                raise RuntimeError("No expiry set on the access token.")
            access_token_expiry = access_token_claims["exp"]
            exp = int(access_token_expiry) - int(current)
            logger.debug(
                f"Token expiry: {exp}  - current is {current} "
                f"and expiry is set to {access_token_expiry} in the token"
            )
            self.general_cache_backend.set(cache_key, access_token_claims, exp)
        return access_token_claims

    def call_validate_tokens_hook(
        self, access_token_jwt: str, client: OIDCClient
    ) -> Any:
        if self.opsettings.get("hook_validate_access_token") is not None:
            logger.debug("OIDC, Calling hook_validate_access_token.")
            return self.call_function(
                "hook_validate_access_token", access_token_jwt, client
            )
        else:
            logger.debug(
                "No way to extract claims from access token. 'use_introspection_on_access_tokens' is false and no user 'hook_validate_access_token' is defined. Empty dict return for access token."
            )
            return None



import contextlib
import logging
import os
import subprocess
import time
from typing import Generator
from unittest.mock import MagicMock, patch

from django.core.servers.basehttp import ThreadedWSGIServer, WSGIRequestHandler
from django.test import override_settings
from django.test.testcases import LiveServerTestCase, LiveServerThread

logger = logging.getLogger(__name__)


class NotReadyException(Exception):
    pass


class OIDCE2ETestCase(LiveServerTestCase):

    docker_id = None
    workdir = None
    docker_workdir = None

    class VerboseLiveServerThread(LiveServerThread):
        def _create_server(self, connections_override=None):
            return ThreadedWSGIServer(
                (self.host, self.port),
                WSGIRequestHandler,
                allow_reuse_address=False,
                connections_override=connections_override,
            )

    server_thread_class = VerboseLiveServerThread


@override_settings(
    ALLOWED_HOSTS=["testserver"],
    STATIC_URL="/static",
    MIDDLEWARE=[
        "django.contrib.sessions.middleware.SessionMiddleware",
        "django.middleware.common.CommonMiddleware",
        "django.contrib.auth.middleware.AuthenticationMiddleware",
        "django.contrib.messages.middleware.MessageMiddleware",
    ],
)
class OIDCE2ELemonLdapNgTestCase(OIDCE2ETestCase):
    @classmethod
    def setUpClass(cls):
        print(" *** Live Server Test Case Setup (LemonLdap::Ng) ***")
        super().setUpClass()

        cls.workdir = os.getcwd()
        print(f"Current workdir: {cls.workdir} ...")
        cls.docker_workdir = f"{cls.workdir}/tests/e2e/docker"
        os.chdir(cls.docker_workdir)
        cls.docker_id = None
        try:
            print("Building LemonLdap docker image...")
            subprocess.run(
                [
                    "docker",
                    "buildx",
                    "build",
                    "-f",
                    "Dockerfile-lemonldap",
                    "-t",
                    "oidc-test-lemonldap-image",
                    ".",
                ]
            )
            print("Running LemonLdap docker image...")
            command = (
                "docker run"
                " --detach --rm -it"
                " -p 8070:8070"
                " -p 8071:8071"
                " -p 8072:8072"
                " -p 8073:8073"
                " -p 8999:9000"
                " -e SSODOMAIN=localhost"
                " -e LOGLEVEL=debug"
                " -e PORTAL_HOSTNAME=localhost:8070"
                " -e MANAGER_HOSTNAME=localhost:8071"
                " -e HANDLER_HOSTNAME=localhost:8072"
                " -e TEST1_HOSTNAME=localhost:8073"
                " -e TEST2_HOSTNAME=localhost.localdomain:8073"
                " oidc-test-lemonldap-image"
            )
            res = subprocess.run(
                command, shell=True, text=True, check=True, capture_output=True
            )
            cls.docker_id = res.stdout.partition("\n")[0]
            print(cls.docker_id)
        except subprocess.CalledProcessError as e:
            print(
                f"Error while launching LemonLdap docker container. errcode: {e.returncode}."
            )
            print(e.stderr)
            if e.returncode == 125:
                print(" +----------|  |--------------------------+ ")
                print(" +---------_|  |_-------------------------+ ")
                print(" +---------\    /-------------------------+ ")  # noqa
                print(" +----------\  /--------------------------+ ")  # noqa
                print(" +-----------\/---------------------------+ ")  # noqa
                print(
                    "   + Try removing any previous LemonLdap image running using this command:"
                )
                print(
                    '   docker stop $(docker ps -a -q --filter ancestor=oidc-test-lemonldap-image --format="{{.ID}}")'
                )
                print(
                    "   + Check also you have no service running on localhost port 8070, 8071, 8072 and 8073."
                )
                print(" +---------------------------------------+ ")
                print(" +---------------------------------------+ ")
                print(" +---------------------------------------+ ")
        finally:
            os.chdir(cls.workdir)
        if cls.docker_id:
            cls.loadLemonLDAPFixtures()
        else:
            raise RuntimeError("Cannot build the context environnement with docker.")

    @classmethod
    def loadLemonLDAPFixtures(cls):
        print(f"Running Django on {cls.live_server_url}")

        print(" + get LemonLDAP info (testing it's up)")
        retry = 0
        ok = False
        while retry < 15 and not ok:
            try:
                cls.docker_lemonldap_command(
                    "/usr/share/lemonldap-ng/bin/lemonldap-ng-cli info"
                )
                ok = True
            except NotReadyException:
                print("   ->  waiting for lemonldap startup...")
                time.sleep(2)
            finally:
                retry += 1

        if retry == 15:
            print(" Aborting, seems LemonLdap-ng is not starting fast enough.")
            raise RuntimeError("Cannot reach lemonldap via cli in time.")

        print(" + Set Global Configuration.")
        cls.docker_lemonldap_command(
            """/usr/share/lemonldap-ng/bin/lemonldap-ng-cli\
    -yes 1 \
    -safe 1\
      set \
        portal http://localhost:8070/ \
        mailUrl http://localhost:8070/resetpwd \
        registerUrl http://localhost:8070/register \
        https 0 \
        securedCookie 0 """
        )

        print(" + Enable OIDC.")
        cls.docker_lemonldap_command(
            """/usr/share/lemonldap-ng/bin/lemonldap-ng-cli \
    -yes 1 \
    -safe 1\
    set \
        issuerDBOpenIDConnectActivation 1 \
        issuerDBOpenIDConnectPath '^/oauth2/' """
        )

        print(" + Create RSA key")
        cls.docker_lemonldap_command(
            """rm -f /tmp/oidc*.key \
    && openssl genrsa -out /tmp/oidc.key 4096 \
    && openssl rsa -pubout \
        -in /tmp/oidc.key \
        -out /tmp/oidc_pub.key \
    && ls -alh /tmp/ \
    && PRIVATE=$(cat /tmp/oidc.key) \
    && PUBLIC=$(cat /tmp/oidc_pub.key) \
    && /usr/share/lemonldap-ng/bin/lemonldap-ng-cli -yes 1 \
      set \
        oidcServicePrivateKeySig "${PRIVATE}" \
    && /usr/share/lemonldap-ng/bin/lemonldap-ng-cli -yes 1 \
      set \
        oidcServicePublicKeySig "${PUBLIC}" \
    && /usr/share/lemonldap-ng/bin/lemonldap-ng-cli -yes 1 \
      set \
        oidcServiceKeyIdSig "somethingsomething" """
        )

        print(" + Create client applications.")
        cls.registerClient(
            "app1",
            "secret_app1",
            cls.live_server_url,
            callback_url=f"{cls.live_server_url}/prefix_lemon1/lemon1-callback",
            post_login_url=f"{cls.live_server_url}/test-ll-logout-done-1",
        )
        cls.registerClient(
            "app1-api",
            "secret_app1-api",
            cls.live_server_url,
            bearerOnly=True,
        )
        cls.registerClient(
            "app2-full",
            "secret_app2-full",
            cls.live_server_url,
            callback_url=f"{cls.live_server_url}/prefix_lemon2/lemon2-callback",
            post_login_url=f"{cls.live_server_url}/test-ll-logout-done-2",
        )
        cls.registerClient(
            "app2-api", "secret_app2-api", cls.live_server_url, bearerOnly=True
        )
        # Default demo users:
        # rtyler :: rtyler
        # msmith :: msmith
        # dwho :: dwho (administrator)

    @classmethod
    def tearDownClass(cls):
        print(" *** Live Server Test Case : Teardown ***")

        # print(" + Debug LemonLdap conf.")
        # conf = cls.docker_lemonldap_command("/usr/share/lemonldap-ng/bin/lemonldap-ng-cli save > /tmp/conf && cat /tmp/conf")
        # print(conf.stdout)

        print("Removing lemonldap docker image...")

        os.chdir(cls.docker_workdir)
        try:
            cmd = (
                "docker stop $("
                'docker ps -a -q --filter ancestor=oidc-test-lemonldap-image --format="{{.ID}}"'
                ")"
            )
            subprocess.run(cmd, shell=True, text=True, check=True)
        except subprocess.CalledProcessError as e:
            print(f"Error stopping lemonldap container: {e.returncode}.")
            print(e.stderr)
        finally:
            os.chdir(cls.workdir)

    @classmethod
    def docker_lemonldap_command(cls, command: str):
        logger.debug(f"Docker LemonLDAP command: {command}")
        # TEMP
        print(f"Docker LemonLDAP command: {command}")
        cmd_prefix = f"docker exec -i {cls.docker_id}"
        # command = command.replace('"', '"')
        final_command = f"{cmd_prefix} /bin/bash <<'EOF'\n{command}\nEOF"
        # print(final_command)
        os.chdir(cls.docker_workdir)
        output = ""
        try:
            output = subprocess.run(
                final_command,
                shell=True,
                text=True,
                check=True,
                capture_output=True,
            )
        except subprocess.CalledProcessError as e:
            errors = e.stderr.split("\n")
            last_error = errors[-1:][0]
            if last_error == "" and len(errors) > 1:
                last_error = errors[-2:-1][0]
            # print(f"Last Error: >{last_error}<")
            print(
                f"Error while launching command on the keycloak container. errcode: {e.returncode}."
            )
            print(e.stderr)
        finally:
            os.chdir(cls.workdir)
        return output

    @classmethod
    def registerClient(
        cls, name, secret, url, callback_url="", post_login_url="", bearerOnly=False
    ):
        redirectUris = "''" if bearerOnly else f"'{callback_url}'"
        logoutRedirectUris = "''" if bearerOnly else f"'{post_login_url}'"

        cls.docker_lemonldap_command(
            f"""usr/share/lemonldap-ng/bin/lemonldap-ng-cli -yes 1 \
    addKey \
        oidcRPMetaDataOptions/{name} oidcRPMetaDataOptionsClientID {name} \
        oidcRPMetaDataOptions/{name} oidcRPMetaDataOptionsClientSecret {secret} \
        oidcRPMetaDataOptions/{name} oidcRPMetaDataOptionsPublic 0 \
        oidcRPMetaDataOptions/{name} oidcRPMetaDataOptionsRedirectUris {redirectUris} \
        oidcRPMetaDataOptions/{name} oidcRPMetaDataOptionsPostLogoutRedirectUris '{logoutRedirectUris}' \
        oidcRPMetaDataOptions/{name} oidcRPMetaDataOptionsIDTokenSignAlg RS512 \
        oidcRPMetaDataOptions/{name} oidcRPMetaDataOptionsIDTokenExpiration 3600 \
        oidcRPMetaDataOptions/{name} oidcRPMetaDataOptionsAccessTokenExpiration 3600 \
        oidcRPMetaDataOptions/{name} oidcRPMetaDataOptionsAllowClientCredentialsGrant 0 \
        oidcRPMetaDataOptions/{name} oidcRPMetaDataOptionsAllowPasswordGrant 0 \
        oidcRPMetaDataOptions/{name} oidcRPMetaDataOptionsoidcRPMetaDataOptionsRefreshToken 1 \
        oidcRPMetaDataOptions/{name} oidcRPMetaDataOptionsLogoutType front \
        oidcRPMetaDataOptions/{name} oidcRPMetaDataOptionsBypassConsent 1 \
           """
        )
        cls.docker_lemonldap_command(
            f""" /usr/share/lemonldap-ng/bin/lemonldap-ng-cli -yes 1 \
    addKey \
        oidcRPMetaDataExportedVars/{name} email mail \
        oidcRPMetaDataExportedVars/{name} family_name sn \
        oidcRPMetaDataExportedVars/{name} name cn \
           """
        )


@override_settings(
    ALLOWED_HOSTS=["testserver"],
    STATIC_URL="/static",
    MIDDLEWARE=[
        "django.contrib.sessions.middleware.SessionMiddleware",
        "corsheaders.middleware.CorsMiddleware",
        "django.middleware.common.CommonMiddleware",
        "django.contrib.auth.middleware.AuthenticationMiddleware",
        "django.contrib.messages.middleware.MessageMiddleware",
    ],
    DJANGO_PYOIDC={
        "sso1": {
            "provider_class": "KeycloakProvider",
            "keycloak_base_uri": "http://localhost:8080/auth/",
            "keycloak_realm": "realm1",
            "client_id": "app1",
            "client_secret": "secret_app1",
            "cache_django_backend": "default",
            "callback_uri_name": "e2e_test_callback_1",
            "post_login_uri_success": "/test-success-1",
            "post_login_uri_failure": "/test-failure-1",
            "post_logout_redirect_uri": "/test-logout-done-1",
            "HOOK_USER_LOGIN": "tests.e2e.test_app.callback:login_callback",
            "HOOK_USER_LOGOUT": "tests.e2e.test_app.callback:logout_callback",
            "login_uris_redirect_allowed_hosts": ["testserver"],
            "login_redirection_requires_https": False,
        },
        # sso2 use a different client_id
        "sso2": {
            "provider_class": "KeycloakProvider",
            "keycloak_base_uri": "http://localhost:8080/auth/",
            "keycloak_realm": "realm1",
            "client_id": "app2-full",
            "client_secret": "secret_app2-full",
            "cache_django_backend": "default",
            "callback_uri_name": "e2e_test_callback_2",
            "post_login_uri_success": "/test-success-2",
            "post_login_uri_failure": "/test-failure-2",
            "post_logout_redirect_uri": "/test-logout-done-2",
            "HOOK_USER_LOGIN": "tests.e2e.test_app.callback:login_callback",
            "HOOK_USER_LOGOUT": "tests.e2e.test_app.callback:logout_callback",
            "login_uris_redirect_allowed_hosts": ["testserver"],
            "login_redirection_requires_https": False,
            "hook_get_user": "tests.e2e.test_app.callback:get_user_with_resource_access_check",
        },
        # broken client_id
        "sso3": {
            "client_id": "bad_client_id",
            "provider_class": "KeycloakProvider",
            "keycloak_base_uri": "http://localhost:8080/auth/",
            "keycloak_realm": "realm1",
            "client_secret": "secret_app1",
            "cache_django_backend": "default",
            "callback_uri_name": "e2e_test_callback_3",
            "post_login_uri_success": "/test-success-3",
            "post_login_uri_failure": "/test-failure-3",
            "post_logout_redirect_uri": "/test-logout-done-3",
            "HOOK_USER_LOGIN": "tests.e2e.test_app.callback:login_callback",
            "HOOK_USER_LOGOUT": "tests.e2e.test_app.callback:logout_callback",
            "login_uris_redirect_allowed_hosts": ["testserver"],
            "login_redirection_requires_https": False,
        },
        # hook_get_user
        "sso4": {
            "provider_class": "KeycloakProvider",
            "keycloak_base_uri": "http://localhost:8080/auth/",
            "keycloak_realm": "realm1",
            "client_id": "app1",
            "client_secret": "secret_app1",
            "cache_django_backend": "default",
            "callback_uri_name": "e2e_test_callback_4",
            "post_login_uri_success": "/test-success-4",
            "post_login_uri_failure": "/test-failure-4",
            "post_logout_redirect_uri": "/test-logout-done-4",
            "login_uris_redirect_allowed_hosts": ["testserver"],
            "login_redirection_requires_https": False,
            "HOOK_USER_LOGIN": "tests.e2e.test_app.callback:login_callback",
            "HOOK_USER_LOGOUT": "tests.e2e.test_app.callback:logout_callback",
            "hook_get_user": "tests.e2e.test_app.callback:get_user_with_resource_access_check",
            "use_introspection_on_access_tokens": True,
        },
        # hook_get_user
        "sso5": {
            "provider_class": "KeycloakProvider",
            "keycloak_base_uri": "http://localhost:8080/auth/",
            "keycloak_realm": "realm1",
            "client_id": "app1",
            "client_secret": "secret_app1",
            "cache_django_backend": "default",
            "callback_uri_name": "e2e_test_callback_5",
            "post_login_uri_success": "/test-success-5",
            "post_login_uri_failure": "/test-failure-5",
            "post_logout_redirect_uri": "/test-logout-done-5",
            "login_uris_redirect_allowed_hosts": ["testserver"],
            "login_redirection_requires_https": False,
            "HOOK_USER_LOGIN": "tests.e2e.test_app.callback:login_callback",
            "HOOK_USER_LOGOUT": "tests.e2e.test_app.callback:logout_callback",
            "use_introspection_on_access_tokens": True,
            "hook_get_user": "tests.e2e.test_app.callback:get_user_with_minimal_audiences_check",
        },
        # API
        "drf": {
            "client_id": "app1-api",
            "cache_django_backend": "default",
            "provider_discovery_uri": "http://localhost:8080/auth/realms/realm1",
            "client_secret": "secret_app1-api",
            "provider_class": "KeycloakProvider",
            "keycloak_base_uri": "http://localhost:8080/auth/",
            "keycloak_realm": "realm1",
        },
    },
)
class OIDCE2EKeycloakTestCase(OIDCE2ETestCase):
    @classmethod
    def setUpClass(cls):
        print(" *** Live Server Test Case Setup (Keycloak) ***")
        super().setUpClass()

        cls.workdir = os.getcwd()
        print(f"Current workdir: {cls.workdir} ...")
        cls.docker_workdir = f"{cls.workdir}/tests/e2e/docker"
        os.chdir(cls.docker_workdir)
        cls.docker_id = None
        try:
            print("Building keycloak docker image...")
            subprocess.run(
                [
                    "docker",
                    "build",
                    "-f",
                    "Dockerfile-keycloak",
                    "-t",
                    "oidc-test-keycloak-image",
                    ".",
                ]
            )
            print("Running keycloak docker image...")
            command = (
                "docker run"
                " --detach --rm -it"
                " -p 8080:8080"
                " -e KEYCLOAK_ADMIN=admin -e KEYCLOAK_ADMIN_PASSWORD=admin"
                " oidc-test-keycloak-image"
            )
            res = subprocess.run(
                command, shell=True, text=True, check=True, capture_output=True
            )
            cls.docker_id = res.stdout.partition("\n")[0]
            print(cls.docker_id)

            cls.docker_front_workdir = f"{cls.workdir}/tests/e2e/front_test_app"
            os.chdir(cls.docker_front_workdir)
            print("Building front docker image...")
            subprocess.run(
                [
                    "docker",
                    "build",
                    "-f",
                    "Dockerfile",
                    "-t",
                    "oidc-test-front-image",
                    ".",
                ]
            )
            command = (
                "docker run"
                " --detach --rm -it"
                " -p 9999:9999"
                f" -e KEYCLOAK_URL=http://localhost:8080/auth -e BACKEND_URL={cls.live_server_url}"
                " oidc-test-front-image"
            )
            subprocess.run(
                command, shell=True, text=True, check=True, capture_output=True
            )
        except subprocess.CalledProcessError as e:
            print(
                f"Error while launching keycloak docker container. errcode: {e.returncode}."
            )
            print(e.stderr)
            if e.returncode == 125:
                print(" +----------|  |--------------------------+ ")
                print(" +---------_|  |_-------------------------+ ")
                print(" +---------\    /-------------------------+ ")  # noqa
                print(" +----------\  /--------------------------+ ")  # noqa
                print(" +-----------\/---------------------------+ ")  # noqa
                print(
                    "   + Try removing any previous Keycloak and front images running using these commands:"
                )
                print(
                    '   docker stop $(docker ps -a -q --filter ancestor=oidc-test-keycloak-image --format="{{.ID}}")'
                )
                print(
                    '   docker stop $(docker ps -a -q --filter ancestor=oidc-test-front-image --format="{{.ID}}")'
                )
                print("   + Check also you have no service running on localhost:8080.")
                print(" +---------------------------------------+ ")
                print(" +---------------------------------------+ ")
                print(" +---------------------------------------+ ")
        finally:
            os.chdir(cls.workdir)
        if cls.docker_id:
            cls.loadKeycloakFixtures()
        else:
            raise RuntimeError("Cannot build the context environnement with docker.")

    @classmethod
    def loadKeycloakFixtures(cls):
        print(f"Running Django on {cls.live_server_url}")

        print(" + connect as Keycloak admin")
        retry = 0
        ok = False
        while retry < 15 and not ok:
            try:
                cls.docker_keycloak_command(
                    "bin/kcadm.sh config credentials --server http://127.0.0.1:8080/auth/ --realm master --user admin --password admin"
                )
                ok = True
            except NotReadyException:
                print("   ->  waiting for keycloak startup...")
                time.sleep(2)
            finally:
                retry += 1

        if retry == 15:
            print(" Aborting, seems Keycloak is not starting fast enough.")
            raise RuntimeError("Cannot reach Keycloak admin in time.")

        print(" + Create test realm")
        cls.docker_keycloak_command(
            "bin/kcadm.sh create realms -s realm=realm1 -s enabled=true"
        )

        print(" + Remove offline authorization default role affectation.")
        cls.docker_keycloak_command(
            "bin/kcadm.sh remove-roles --rname default-roles-realm1 --rolename offline_access -r realm1"
        )

        print(" + Create client applications.")
        app1_id = cls.registerClient(
            "app1",
            "secret_app1",
            cls.live_server_url,
            serviceAccount=False,
            channelLogoutUrl=f"{cls.live_server_url}/back_channel_logout-1/",
        )
        app1_api_id = cls.registerClient(
            "app1-api",
            "secret_app1-api",
            cls.live_server_url,
            bearerOnly=True,
            serviceAccount=False,
        )
        app1_front_id = cls.registerClient(
            "app1-front",
            None,
            "http://localhost:9999",
            bearerOnly=False,
            serviceAccount=False,
            channelLogoutUrl="http://localhost:9999",
        )
        app_m2m1_id = cls.registerClient(
            "app_m2m1",
            "secret_app-m2m1",
            cls.live_server_url,
            bearerOnly=False,
            serviceAccount=True,
        )
        app2_m2m2_id = cls.registerClient(
            "app2_m2m2",
            "secret_app2-m2m2",
            cls.live_server_url,
            bearerOnly=False,
            serviceAccount=True,
        )
        app2_full_id = cls.registerClient(
            "app2-full",
            "secret_app2-full",
            cls.live_server_url,
            bearerOnly=False,
            channelLogoutUrl=f"{cls.live_server_url}/back_channel_logout-2/",
        )
        app2_api_id = cls.registerClient(
            "app2-api", "secret_app2-api", cls.live_server_url, bearerOnly=True
        )

        print(" + Create client applications access roles.")
        app1_role = cls.registerClientRole(app1_id, "AccessApp1")
        app1_bis_role = cls.registerClientRole(app1_api_id, "AccessApp1API")
        app1_front_role = cls.registerClientRole(app1_front_id, "AccessApp1Front")
        app2_full_role = cls.registerClientRole(app2_full_id, "AccessApp2Full")
        app2_api_role = cls.registerClientRole(app2_api_id, "AccessApp2API")

        print(" + Create Client Scopes.")
        id_zone_app1 = cls.registerClientScope(
            "zone-app1",
            [
                {app1_id: app1_role},
                {app1_api_id: app1_bis_role},
                {app1_front_id: app1_front_role},
            ],
        )
        id_zone_app2 = cls.registerClientScope(
            "zone-app2",
            [
                {app2_full_id: app2_full_role},
                {app2_api_id: app2_api_role},
            ],
        )

        print(" + Update applications client scopes")
        cls.addClientScopeForClient(app1_id, id_zone_app1)
        cls.addClientScopeForClient(app1_api_id, id_zone_app1)
        cls.addClientScopeForClient(app1_front_id, id_zone_app1)
        cls.addClientScopeForClient(app_m2m1_id, id_zone_app1)
        cls.addClientScopeForClient(app2_full_id, id_zone_app2)
        cls.addClientScopeForClient(app2_api_id, id_zone_app2)
        cls.addClientScopeForClient(app2_m2m2_id, id_zone_app2)

        print(" + Create Groups.")
        gApp1 = cls.registerGroup(
            "App1",
            [
                {"app1": "AccessApp1"},
                {"app1-api": "AccessApp1API"},
                {"app1-front": "AccessApp1Front"},
            ],
        )
        gApp2 = cls.registerGroup(
            "App2",
            [
                {"app2-full": "AccessApp2Full"},
                {"app2-api": "AccessApp2API"},
            ],
        )
        gApp1Restricted = cls.registerGroup(
            "App1Restricted",
            [
                {"app1": "AccessApp1"},
            ],
        )
        gm2m = cls.registerGroup(
            "GroupM2M",
            [
                {"app1": "AccessApp1"},
                {"app1-api": "AccessApp1API"},
                {"app2-full": "AccessApp2Full"},
                {"app2-api": "AccessApp2API"},
            ],
        )
        gAppAll = cls.registerGroup(
            "AllApps",
            [
                {"app1": "AccessApp1"},
                {"app1-api": "AccessApp1API"},
                {"app1-front": "AccessApp1Front"},
                {"app2-full": "AccessApp2Full"},
                {"app2-api": "AccessApp2API"},
            ],
        )
        print(" + Link service account users to groups")
        m2m_user1 = cls.searchUser("service-account-app_m2m1")
        m2m_user2 = cls.searchUser("service-account-app2_m2m2")
        cls.add_user_to_group(m2m_user1, gm2m)
        cls.add_user_to_group(m2m_user2, gm2m)

        print(" + Create users.")
        cls.registerUser(
            "user1",
            "passwd1",
            groups=[
                gAppAll,
            ],
        )
        cls.registerUser(
            "user_limit_app2",
            "passwd2",
            groups=[
                gApp2,
            ],
        )
        cls.registerUser(
            "user_limit_app1",
            "passwd1",
            groups=[
                gApp1,
            ],
        )
        cls.registerUser(
            "user_app1_only",
            "passwd1",
            groups=[gApp1Restricted],
        )

    @classmethod
    def tearDownClass(cls):
        print(" *** Live Server Test Case : Teardown ***")

        print("Extracting keycloak logs before stopping...")
        os.chdir(cls.docker_workdir)
        try:
            subprocess.run(
                f"docker logs {cls.docker_id}", shell=True, text=True, check=True
            )
        except subprocess.CalledProcessError as e:
            print(f"Error while retrieving keycloak docker logs: {e.returncode}.")
            print(e.stderr)
        finally:
            os.chdir(cls.workdir)

        print("Removing keycloak docker image...")

        os.chdir(cls.docker_workdir)
        try:
            cmd = (
                "docker stop $("
                'docker ps -a -q --filter ancestor=oidc-test-keycloak-image --format="{{.ID}}"'
                ")"
            )
            subprocess.run(cmd, shell=True, text=True, check=True)
        except subprocess.CalledProcessError as e:
            print(f"Error stopping keycloak container: {e.returncode}.")
            print(e.stderr)
        finally:
            os.chdir(cls.docker_front_workdir)
            print("Removing front docker image...")
            try:
                cmd = (
                    "docker stop $("
                    'docker ps -a -q --filter ancestor=oidc-test-front-image --format="{{.ID}}"'
                    ")"
                )
                subprocess.run(cmd, shell=True, text=True, check=True)
            except subprocess.CalledProcessError as e:
                print(f"Error stopping front container: {e.returncode}.")
                print(e.stderr)
            finally:
                os.chdir(cls.workdir)

    @classmethod
    def docker_keycloak_command(cls, command: str):
        logger.debug(f"Docker Keycloak command: {command}")
        # print(f"Docker Keycloak command: {command}")
        cmd_prefix = f"docker exec {cls.docker_id}"
        command = command.replace('"', '"')
        final_command = f"{cmd_prefix} /bin/bash -c '{command}'"
        # print(final_command)
        os.chdir(cls.docker_workdir)
        output = ""
        try:
            output = subprocess.run(
                final_command,
                shell=True,
                text=True,
                check=True,
                capture_output=True,
            )
        except subprocess.CalledProcessError as e:
            errors = e.stderr.split("\n")
            last_error = errors[-1:][0]
            if last_error == "" and len(errors) > 1:
                last_error = errors[-2:-1][0]
            # print(f"Last Error: >{last_error}<")
            if e.returncode == 1 and last_error in [
                "Failed to send request - Connect to 127.0.0.1:8080 [/127.0.0.1] failed: Connection refused",
                "Failed to send request - Connect to 127.0.0.1:8080 [/127.0.0.1] failed: Connection refused (Connection refused)",
                "Invalid user credentials [invalid_grant]",
            ]:
                raise NotReadyException()
            else:
                print(
                    f"Error while launching command on the keycloak container. errcode: {e.returncode}."
                )
                print(e.stderr)
        finally:
            os.chdir(cls.workdir)
        return output

    @classmethod
    def add_user_to_group(cls, user_id, group):
        cls.docker_keycloak_command(
            f"bin/kcadm.sh update users/{user_id}/groups/{group} -r realm1"
        )

    @classmethod
    def searchUser(cls, user_name):
        output = cls.docker_keycloak_command(
            "bin/kcadm.sh get users -r realm1"
            f" -q username={user_name} --fields 'id,username'"
        )
        id = None
        cpt = 0
        prevline = ""
        for line in output.stdout.splitlines():
            if line == f'  "username" : "{user_name}"':
                cpt = cpt - 1
                id_line = prevline
                # line looks like '    "id": "42562-..-45646",'
                # we split on " and get 3rd value
                id = id_line.split('"')[3]
                break
            cpt += 1
            prevline = line
        return id

    @classmethod
    def registerClient(
        cls,
        name,
        secret,
        url,
        bearerOnly=False,
        serviceAccount=False,
        channelLogoutUrl="",
    ):
        if serviceAccount:
            bServiceAccountEnabled = "true"
            redirectUris = "[ ]"
            extraAttributes = (
                '"use.refresh.tokens": "true",'
                '"client_credentials.use_refresh_token": "false",'
                '"tls.client.certificate.bound.access.tokens": "false",'
                '"require.pushed.authorization.requests": "false",'
                '"acr.loa.map": "{}",'
                '"token.response.type.bearer.lower-case": "false",'
            )
            backchannelLogoutUrl = ""
            frontchannelLogoutUrl = ""
        else:

            bServiceAccountEnabled = "false"
            redirectUris = "[ ]" if bearerOnly else f'[ "{url}/*" ]'
            extraAttributes = ""
        bBearerOnly = "true" if bearerOnly else "false"
        bStandardFlowEnabled = "false" if bearerOnly else "true"
        if secret is None:
            public_line = '"publicClient" : true,'
            secret_line = ""
            frontch_line = '"frontchannelLogout" : true,'
            backchannelLogoutUrl = ""
            frontchannelLogoutUrl = channelLogoutUrl
        else:
            public_line = '"publicClient" : false,'
            frontch_line = '"frontchannelLogout" : false,'
            frontchannelLogoutUrl = ""
            backchannelLogoutUrl = channelLogoutUrl
            secret_line = (
                f'"clientAuthenticatorType" : "client-secret", "secret" : "{secret}",'
            )
        output = cls.docker_keycloak_command(
            f"""bin/kcadm.sh create clients -r realm1 -f - << EOF
{{
    "clientId": "{name}",
    "name": "{name}",
    "description": "{name}",
    "rootUrl": "{url}",
    "adminUrl" : "",
    "baseUrl" : "",
    "surrogateAuthRequired" : false,
    "enabled" : true,
    "alwaysDisplayInConsole" : false,
    {secret_line}
    "redirectUris" : {redirectUris},
    "webOrigins" : [ "+" ],
    "notBefore" : 0,
    "bearerOnly" : {bBearerOnly},
    "consentRequired" : false,
    "standardFlowEnabled" : {bStandardFlowEnabled},
    "implicitFlowEnabled" : false,
    "directAccessGrantsEnabled" : false,
    "serviceAccountsEnabled" : {bServiceAccountEnabled},
    {public_line}
    {frontch_line}
    "protocol" : "openid-connect",
    "attributes" : {{
      {extraAttributes}
      "oidc.ciba.grant.enabled" : "false",
      "post.logout.redirect.uris" : "+",
      "display.on.consent.screen" : "false",
      "oauth2.device.authorization.grant.enabled" : "false",
      "backchannel.logout.revoke.offline.tokens" : "false",
      "backchannel.logout.session.required" : "false",
      "backchannel.logout.url": "{backchannelLogoutUrl}",
      "frontchannel.logout.url": "{frontchannelLogoutUrl}"
    }},
    "authenticationFlowBindingOverrides" : {{ }},
    "fullScopeAllowed" : false,
    "defaultClientScopes" : [ "web-origins", "acr", "roles", "profile", "email" ],
    "optionalClientScopes" : [ "address", "phone" ]
}}
EOF"""
        )
        # output was in the form
        # "Created new client with id '0f5d0645-a7a7-4e88-adf8-b9e568025f5c'""
        # we need to extract the id from this message
        id = None
        if output.stderr[:18] == "Created new client":
            id = output.stderr.split(" ")[5].strip("'\r\n")
        return id

    @classmethod
    def addClientScopeForClient(cls, client_id, clientscope_id):
        cls.docker_keycloak_command(
            "bin/kcadm.sh update"
            f" clients/{client_id}/default-client-scopes/{clientscope_id}"
            " -r realm1"
        )

    @classmethod
    def registerClientRole(cls, client_id, role):
        output = cls.docker_keycloak_command(
            "bin/kcadm.sh create "
            f"clients/{client_id}/roles"
            " -r realm1"
            f" -s name={role}"
            " -o --fields id,name,composite,clientRole,containerId"
        )
        return output.stdout

    @classmethod
    def registerClientScope(cls, client_scope, role_mappings=[]):
        output = cls.docker_keycloak_command(
            f"""bin/kcadm.sh create client-scopes -r realm1 -f - << EOF
{{
    "name": "{client_scope}",
    "protocol" : "openid-connect",
    "description" : "{client_scope}",
    "attributes" : {{
      "include.in.token.scope" : "false",
      "display.on.consent.screen" : "false"
    }}
}}
EOF"""
        )
        # output was in the form
        # "Created new client-scope with id '0f5d0645-a7a7-4e88-adf8-b9e568025f5c'""
        # we need to extract the id from this message
        id = None
        if output.stderr[:18] == "Created new client":
            id = output.stderr.split(" ")[5].strip("'\r\n")
        if id:
            for role_mapping in role_mappings:
                for client_id, role_definition in role_mapping.items():
                    cls.docker_keycloak_command(
                        f"""bin/kcadm.sh create client-scopes/{id}/scope-mappings/clients/{client_id} -r realm1 -f - << EOF
[{role_definition}]
EOF"""
                    )
        return id

    @classmethod
    def registerGroup(cls, name, role_mappings=[]):
        output = cls.docker_keycloak_command(
            "bin/kcadm.sh create groups -r realm1" f" -s name={name}"
        )
        # output was in the form
        # "Created new group with id '0f5d0645-a7a7-4e88-adf8-b9e568025f5c'""
        # we need to extract the id from this message
        id = None
        if output.stderr[:17] == "Created new group":
            id = output.stderr.split(" ")[5].strip("'\r\n")
        if id:
            for role_mapping in role_mappings:
                for client_name, role_name in role_mapping.items():
                    cls.docker_keycloak_command(
                        "bin/kcadm.sh add-roles"
                        " -r realm1"
                        f" --gid {id}"
                        f" --cclientid {client_name}"
                        f" --rolename {role_name}"
                    )
        return id

    @classmethod
    def registerUser(cls, user, password, groups=[]):
        output = cls.docker_keycloak_command(
            "bin/kcadm.sh create users"
            " -r realm1"
            f" -s username={user}"
            f" -s email={user}@example.com"
            " -s enabled=true -s emailVerified=true"
        )
        # output was in the form
        # "Created new user with id '0f5d0645-a7a7-4e88-adf8-b9e568025f5c'""
        # we need to extract the id from this message
        user_id = None
        if output.stderr[:16] == "Created new user":
            user_id = output.stderr.split(" ")[5].strip("'\r\n")
        if user_id:
            cls.docker_keycloak_command(
                "bin/kcadm.sh set-password"
                " -r realm1"
                f" --username {user}"
                f" --new-password={password}"
            )
            for group in groups:
                cls.docker_keycloak_command(
                    f"bin/kcadm.sh update users/{user_id}/groups/{group}" " -r realm1"
                )


@contextlib.contextmanager
def wrap_class(obj: object, method: str) -> Generator[MagicMock, None, None]:
    mock = MagicMock()
    real_method = getattr(obj, method)

    def wrap_method(self, *args, **kwargs):
        mock.__call__(*args, **kwargs)
        return real_method(self, *args, **kwargs)

    with patch.object(obj, method, wrap_method):
        yield mock



from unittest import mock

from django.test import TestCase

from django_pyoidc.drf.schema import OIDCScheme


class OIDCShemaTestCase(TestCase):

    @mock.patch("django_pyoidc.drf.schema.OIDCSettingsFactory.get")
    def test_security_definition_no_well_known_endpoint(self, mocked_get_settings):
        mocked_get_settings.return_value = {
            "provider_discovery_uri": "https://sso.demo/realms/test"
        }
        result = OIDCScheme.get_security_definition(None)
        self.assertEqual(
            result["openIdConnectUrl"],
            "https://sso.demo/realms/test/.well-known/openid-configuration",
        )

    @mock.patch("django_pyoidc.drf.schema.OIDCSettingsFactory.get")
    def test_security_definition_with_well_known_endpoint(self, mocked_get_settings):
        mocked_get_settings.return_value = {
            "provider_discovery_uri": "https://sso.demo/realms/test/.well-known/openid-configuration"
        }
        result = OIDCScheme.get_security_definition(None)
        self.assertEqual(
            result["openIdConnectUrl"],
            "https://sso.demo/realms/test/.well-known/openid-configuration",
        )



from django.urls import include, path

from django_pyoidc.helper import OIDCHelper
from django_pyoidc.views import (
    OIDCBackChannelLogoutView,
    OIDCCallbackView,
    OIDCLoginView,
    OIDCLogoutView,
)

helper10 = OIDCHelper(op_name="sso10")
helper11 = OIDCHelper(op_name="sso11")
helper12 = OIDCHelper(op_name="sso12")
helper13 = OIDCHelper(op_name="sso13")
urlpatterns = [
    path("login-xyz/", OIDCLoginView.as_view(op_name="sso1"), name="test_login"),
    path("login-wtf/", OIDCLoginView.as_view(op_name="sso2"), name="test_login_2"),
    path(
        "callback-xyz/",
        OIDCCallbackView.as_view(op_name="sso1"),
        name="my_test_callback",
    ),
    path(
        "callback-wtf/",
        OIDCCallbackView.as_view(op_name="sso2"),
        name="my_test_callback_sso2",
    ),
    path(
        "logout-xyz/",
        OIDCLogoutView.as_view(op_name="sso1"),
        name="test_logout",
    ),
    path(
        "back_channel_logout-xyz/",
        OIDCBackChannelLogoutView.as_view(op_name="sso1"),
        name="test_blogout",
    ),
    path(
        "",
        include(
            (helper10.get_urlpatterns(), "django_pyoidc"), namespace="op10_namespace"
        ),
    ),
    path(
        "prefix11/",
        include(
            (helper11.get_urlpatterns(), "django_pyoidc"), namespace="op11_namespace"
        ),
    ),
    path(
        "prefix12/",
        include(
            (helper12.get_urlpatterns(), "django_pyoidc"), namespace="op12_namespace"
        ),
    ),
    path(
        "prefix13/",
        include(
            (helper13.get_urlpatterns(), "django_pyoidc"), namespace="op13_namespace"
        ),
    ),
]



import base64
import logging
from typing import Any, Dict, List, MutableMapping, Union

import jsonpickle  # type: ignore[import-untyped]
from Cryptodome.PublicKey.RSA import RsaKey, import_key
from django.core.cache import BaseCache, caches
from jsonpickle.handlers import BaseHandler  # type: ignore[import-untyped]
from oic.utils.session_backend import SessionBackend

from django_pyoidc.models import OIDCSession
from django_pyoidc.settings import OIDCSettings

logger = logging.getLogger(__name__)


# From https://github.com/alehuo/pyoidc-redis-session-backend/blob/master/pyoidc_redis_session_backend/__init__.py
class RSAKeyHandler(BaseHandler):  # type: ignore
    def flatten(
        self, obj: RsaKey, data: MutableMapping[str, Any]
    ) -> MutableMapping[str, Any]:
        data["rsa_key"] = base64.b64encode(obj.export_key()).decode("utf-8")
        return data

    def restore(self, obj: Dict[str, str]) -> RsaKey:
        return import_key(base64.b64decode(obj["rsa_key"]))


jsonpickle.register(RsaKey, RSAKeyHandler)


class OIDCCacheSessionBackendForDjango(SessionBackend):
    """Implement Session backend using django cache."""

    def __init__(self, opsettings: OIDCSettings):
        cache_key: str = opsettings.get("cache_django_backend")  # type: ignore[assignment] # we can assume that the configuration is right
        self.storage: BaseCache = caches[cache_key]
        self.op_name = opsettings.get("op_name")

    def get_key(self, key: str) -> str:
        return f"{self.op_name}-{key}"

    def __setitem__(self, key: str, value: Dict[str, Union[str, bool]]) -> None:
        self.storage.set(self.get_key(key), jsonpickle.encode(value))

    def __getitem__(self, key: str) -> Dict[str, Union[str, bool]]:
        data = self.storage.get(self.get_key(key))
        if data is None:
            raise KeyError  # Makes __getItem__ handle like Python dict
        return jsonpickle.decode(data)

    def __delitem__(self, key: str) -> None:
        self.storage.delete(self.get_key(key))

    def __contains__(self, key: str) -> bool:
        return self.storage.get(self.get_key(key)) is not None

    def get_by_uid(self, uid: str) -> List[str]:
        # FIXME : maybe .filter(cache_session_key=uid) ?
        result = OIDCSession.objects.filter(cache_session_key=uid).values_list(
            "cache_session_key", flat=True
        )
        logger.debug(f"Fetched the following sid : {result} for {uid=}")

        return list(result)

    def get_by_sub(self, sub: str) -> List[str]:
        result = OIDCSession.objects.filter(sub=sub).values_list(
            "cache_session_key", flat=True
        )
        logger.debug(f"Fetched fhe following sid : {result} for {sub=}")
        return list(result)

    def get(self, attr: str, val: str) -> List[str]:
        logger.debug(f"Fetch SID for sessions where [{attr}] = {val}")
        raise NotImplementedError(
            "Current session implementation does not support this method"
        )
